# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class DatasetType(Enum):
    DB_TABLE = "DB_TABLE"
    STREAM = "STREAM"
    HTTP_ENDPOINT = "HTTP_ENDPOINT"


class JobType(Enum):
    BATCH = "BATCH"
    STREAM = "STREAM"
    SERVICE = "SERVICE"


class RunState(Enum):
    NEW = 'NEW'
    RUNNING = 'RUNNING'
    COMPLETED = 'COMPLETED'
    FAILED = 'FAILED'
    ABORTED = 'ABORTED'


class DatasetFieldType(Enum):
    NUMBER = 1
    DECIMAL = 2
    NUMERIC = 3
    INT = 4
    INTEGER = 5
    BIGINT = 6
    SMALLINT = 7
    FLOAT = 8
    FLOAT4 = 9
    FLOAT8 = 10
    DOUBLE = 11
    REAL = 12
    VARCHAR = 13
    CHAR = 14
    CHARACTER = 15
    STRING = 16
    TEXT = 17
    BINARY = 18
    VARBINARY = 19
    BOOLEAN = 20
    DATE = 21
    DATETIME = 22
    TIME = 23
    TIMESTAMP = 24
    TIMESTAMP_LTZ = 25
    TIMESTAMP_NTZ = 26
    TIMESTAMP_TZ = 27
    VARIANT = 28
    OBJECT = 29
    ARRAY = 30



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import requests

from marquez_client import errors
from marquez_client.backend import Backend

log = logging.getLogger(__name__)


class HttpBackend(Backend):
    def __init__(self, url, timeout):
        self._timeout = timeout
        self._url = url

    def put(self, path, headers, payload):
        log.debug("_put()")

        response = requests.put(
            url=f'{self._url}{path}', headers=headers, json=payload,
            timeout=self._timeout)

        return self._response(response, as_json=True)

    def post(self, path, headers, payload=None):
        log.debug("_post()")

        response = requests.post(
            url=f'{self._url}{path}', headers=headers, json=payload,
            timeout=self._timeout)

        return self._response(response, as_json=True)

    def _response(self, response, as_json):
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            self._raise_api_error(e)

        return response.json() if as_json else response.text

    def _raise_api_error(self, e):
        # TODO: https://github.com/MarquezProject/marquez-python/issues/55
        raise errors.APIError() from e



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
import logging.config
import random
import unittest
import uuid

import yaml

from marquez_client import Clients
from marquez_client.models import (DatasetType, JobType)

from marquez_client.utils import Utils

_NAMESPACE = 'default'

log = logging.getLogger(__name__)


class TestAirflowDAG(unittest.TestCase):
    def setUp(self):
        log.debug("TestAirflowDAG.setup(): ")

        with open('tests/logConfig.yaml', 'rt') as file:
            yamlConfig = yaml.safe_load(file.read())
            logging.config.dictConfig(yamlConfig)
            log.info("loaded logConfig.yaml")

        os.environ['MARQUEZ_BACKEND'] = 'file'

        self.client = Clients.new_write_only_client()
        log.info("created marquez_client.")

    def test_create_dag(self):
        log.debug("TestAirflowDAG::test_create_dag")

        for i in range(1000):
            NAMESPACE = "my-namespace"
            OWNER = "me"
            SOURCE = "my-source"
            DATASET = f'my-dataset-{i}'
            PHYSICAL = f'public.my_table-{i}'
            run_id = str(uuid.uuid4())
            JOB = f'my-job-{i%10}'

            self.client.create_namespace(NAMESPACE, OWNER)
            self.client.create_source(
                SOURCE,
                'POSTGRESQL',
                "jdbc:postgresql://localhost:5432/test?user=fred&ssl=true")
            self.client.create_dataset(
                NAMESPACE, DATASET, DatasetType.DB_TABLE,
                PHYSICAL, SOURCE, run_id)
            self.client.create_job(NAMESPACE, JOB, JobType.BATCH)
            self.client.create_job_run(NAMESPACE, JOB, run_id,
                                       mark_as_running=True)

            udiff = (i % 10 - random.randrange(10))

            if udiff >= -1 or udiff <= 1:
                self.client.mark_job_run_as_failed(run_id, Utils.utc_now())
            else:
                self.client.mark_job_run_as_completed(run_id, Utils.utc_now())


if __name__ == '__main__':
    unittest.main()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import logging

from six.moves.urllib.parse import quote

from marquez_client.version import VERSION
from .models import (DatasetType, JobType)
from .utils import Utils

_API_PATH = '/api/v1'
_USER_AGENT = f'marquez-python/{VERSION}'
_HEADERS = {'User-Agent': _USER_AGENT}

log = logging.getLogger(__name__)


# Marquez Write Only Client
class MarquezWriteOnlyClient(object):
    def __init__(self, backend):
        self._backend = backend

    # Namespace API
    def create_namespace(self, namespace_name, owner_name, description=None):
        Utils.check_name_length(namespace_name, 'namespace_name')
        Utils.check_name_length(owner_name, 'owner_name')

        payload = {
            'ownerName': owner_name
        }

        if description:
            payload['description'] = description

        return self._backend.put(
            self._path('/namespaces/{0}', namespace_name),
            headers=_HEADERS,
            payload=payload
        )

    # Source API
    def create_source(self, source_name, source_type, connection_url,
                      description=None):
        Utils.check_name_length(source_name, 'source_name')

        Utils.is_valid_connection_url(connection_url)

        payload = {
            'type': source_type,
            'connectionUrl': connection_url
        }

        if description:
            payload['description'] = description

        return self._backend.put(
            self._path('/sources/{0}', source_name),
            headers=_HEADERS,
            payload=payload)

    # Datasets API
    def create_dataset(self, namespace_name, dataset_name, dataset_type,
                       physical_name, source_name, run_id=None,
                       description=None, schema_location=None,
                       fields=None, tags=None):
        Utils.check_name_length(namespace_name, 'namespace_name')
        Utils.check_name_length(dataset_name, 'dataset_name')
        Utils.is_instance_of(dataset_type, DatasetType)

        if dataset_type == DatasetType.STREAM:
            Utils.is_none(schema_location, 'schema_location')

        Utils.check_name_length(physical_name, 'physical_name')
        Utils.check_name_length(source_name, 'source_name')

        payload = {
            'type': dataset_type.value,
            'physicalName': physical_name,
            'sourceName': source_name,
        }

        if description:
            payload['description'] = description

        if run_id:
            payload['runId'] = run_id

        if fields:
            payload['fields'] = fields

        if tags:
            payload['tags'] = tags

        if schema_location:
            payload['schemaLocation'] = schema_location

        return self._backend.put(
            self._path('/namespaces/{0}/datasets/{1}', namespace_name,
                       dataset_name),
            headers=_HEADERS,
            payload=payload
        )

    # Job API
    def create_job(self, namespace_name, job_name, job_type,
                   location=None, input_dataset=None,
                   output_dataset=None, description=None, context=None,
                   run_id=None):
        Utils.check_name_length(namespace_name, 'namespace_name')
        Utils.check_name_length(job_name, 'job_name')
        Utils.is_instance_of(job_type, JobType)

        payload = {
            'inputs': input_dataset or [],
            'outputs': output_dataset or [],
            'type': job_type.name
        }

        if run_id:
            payload['runId'] = run_id

        if context:
            payload['context'] = context

        if location:
            payload['location'] = location

        if description:
            payload['description'] = description

        return self._backend.put(
            self._path('/namespaces/{0}/jobs/{1}', namespace_name, job_name),
            headers=_HEADERS,
            payload=payload
        )

    def create_job_run(self, namespace_name, job_name, run_id,
                       nominal_start_time=None,
                       nominal_end_time=None, run_args=None,
                       mark_as_running=False):
        Utils.check_name_length(namespace_name, 'namespace_name')
        Utils.check_name_length(job_name, 'job_name')

        payload = {}

        payload['id'] = run_id

        if nominal_start_time:
            payload['nominalStartTime'] = nominal_start_time

        if nominal_end_time:
            payload['nominalEndTime'] = nominal_end_time

        if run_args:
            payload['args'] = run_args

        response = self._backend.post(
            self._path('/namespaces/{0}/jobs/{1}/runs',
                       namespace_name, job_name),
            headers=_HEADERS,
            payload=payload)

        if mark_as_running:
            response = self.mark_job_run_as_started(
                run_id, str(datetime.datetime.utcnow()))

        return response

    def mark_job_run_as_started(self, run_id, action_at=None):
        return self.__mark_job_run_as(run_id, 'start', action_at)

    def mark_job_run_as_completed(self, run_id, action_at=None):
        return self.__mark_job_run_as(run_id, 'complete', action_at)

    def mark_job_run_as_failed(self, run_id, action_at=None):
        return self.__mark_job_run_as(run_id, 'fail', action_at)

    def mark_job_run_as_aborted(self, run_id, action_at=None):
        return self.__mark_job_run_as(run_id, 'abort', action_at)

    def __mark_job_run_as(self, run_id, action, action_at=None):
        Utils.is_valid_uuid(run_id, 'run_id')

        return self._backend.post(
            self._path('/jobs/runs/{0}/{1}?at={2}', run_id, action,
                       action_at if action_at else Utils.utc_now()),
            headers=_HEADERS,
            payload={}
        )

    # Common
    @staticmethod
    def _path(path_template, *args):
        encoded_args = [quote(arg.encode('utf-8'), safe='') for arg in args]
        return f'{_API_PATH}{path_template.format(*encoded_args)}'



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os

from marquez_client.backend import Backend

log = logging.getLogger(__name__)


class FileBackend(Backend):
    def __init__(self, file):
        self._file = open(self._mkdir_if_not_exists(file), 'a')

    @staticmethod
    def _mkdir_if_not_exists(file):
        path, _ = os.path.split(file)
        os.makedirs(path, exist_ok=True)
        return file

    def put(self, path, headers, payload):
        log.debug("_put()")

        put_details = {}

        put_details['method'] = 'PUT'
        put_details['path'] = path
        put_details['headers'] = headers
        put_details['payload'] = payload

        put_json = json.dumps(put_details)
        log.info(put_json)

        self._sync_file(put_json)

    def post(self, path, headers, payload=None):
        log.debug("_post()")

        post_details = {}

        post_details['method'] = 'POST'
        post_details['path'] = path
        post_details['headers'] = headers
        if payload:
            post_details['payload'] = payload

        post_json = json.dumps(post_details)
        log.info(post_json)

        self._sync_file(post_json)

    def _sync_file(self, json_data):
        self._file.write(json_data)
        self._file.write(os.linesep)
        self._file.flush()
        os.fsync(self._file.fileno())

    def __del__(self):
        self._file.close()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

import requests
from six.moves.urllib.parse import quote

from marquez_client import errors
from marquez_client.constants import (DEFAULT_TIMEOUT_MS)
from marquez_client.models import (DatasetType, JobType)
from marquez_client.utils import Utils
from marquez_client.version import VERSION

_API_PATH = '/api/v1'
_USER_AGENT = f'marquez-python/{VERSION}'
_HEADERS = {'User-Agent': _USER_AGENT}

log = logging.getLogger(__name__)


# Marquez Client
class MarquezClient(object):
    def __init__(self, url, timeout_ms=None):
        self._timeout = Utils.to_seconds(timeout_ms or os.environ.get(
            'MARQUEZ_TIMEOUT_MS', DEFAULT_TIMEOUT_MS)
        )

        self._api_base = f'{url}{_API_PATH}'

        log.debug(self._api_base)

    # Namespace API
    def create_namespace(self, namespace_name, owner_name, description=None):
        Utils.check_name_length(namespace_name, 'namespace_name')
        Utils.check_name_length(owner_name, 'owner_name')

        payload = {
            'ownerName': owner_name
        }

        if description:
            payload['description'] = description

        return self._put(
            self._url('/namespaces/{0}', namespace_name),
            payload=payload
        )

    def get_namespace(self, namespace_name):
        Utils.check_name_length(namespace_name, 'namespace_name')

        return self._get(self._url('/namespaces/{0}', namespace_name))

    def list_namespaces(self, limit=None, offset=None):
        return self._get(
            self._url('/namespaces'),
            params={
                'limit': limit,
                'offset': offset
            }
        )

    # Source API
    def create_source(self, source_name, source_type, connection_url,
                      description=None):
        Utils.check_name_length(source_name, 'source_name')

        Utils.is_valid_connection_url(connection_url)

        payload = {
            'type': source_type,
            'connectionUrl': connection_url
        }

        if description:
            payload['description'] = description

        return self._put(self._url('/sources/{0}', source_name),
                         payload=payload)

    def get_source(self, source_name):
        Utils.check_name_length(source_name, 'source_name')

        return self._get(self._url('/sources/{0}', source_name))

    def list_sources(self, limit=None, offset=None):
        return self._get(
            self._url('/sources'),
            params={
                'limit': limit,
                'offset': offset
            }
        )

    # Datasets API
    def create_dataset(self, namespace_name, dataset_name, dataset_type,
                       physical_name, source_name,
                       description=None, run_id=None,
                       schema_location=None,
                       fields=None, tags=None):
        Utils.check_name_length(namespace_name, 'namespace_name')
        Utils.check_name_length(dataset_name, 'dataset_name')
        Utils.is_instance_of(dataset_type, DatasetType)

        if dataset_type == DatasetType.STREAM:
            MarquezClient._is_none(schema_location, 'schema_location')

        Utils.check_name_length(physical_name, 'physical_name')
        Utils.check_name_length(source_name, 'source_name')

        payload = {
            'type': dataset_type.value,
            'physicalName': physical_name,
            'sourceName': source_name,
        }

        if description:
            payload['description'] = description

        if run_id:
            payload['runId'] = run_id

        if fields:
            payload['fields'] = fields

        if tags:
            payload['tags'] = tags

        if schema_location:
            payload['schemaLocation'] = schema_location

        return self._put(
            self._url('/namespaces/{0}/datasets/{1}', namespace_name,
                      dataset_name),
            payload=payload
        )

    def get_dataset(self, namespace_name, dataset_name):
        Utils.check_name_length(namespace_name, 'namespace_name')
        Utils.check_name_length(dataset_name, 'dataset_name')

        return self._get(
            self._url('/namespaces/{0}/datasets/{1}',
                      namespace_name, dataset_name)
        )

    def list_datasets(self, namespace_name, limit=None, offset=None):
        Utils.check_name_length(namespace_name, 'namespace_name')

        return self._get(
            self._url('/namespaces/{0}/datasets', namespace_name),
            params={
                'limit': limit,
                'offset': offset
            }
        )

    def tag_dataset(self, namespace_name, dataset_name, tag_name):
        Utils.check_name_length(namespace_name, 'namespace_name')
        Utils.check_name_length(dataset_name, 'dataset_name')

        if not tag_name:
            raise ValueError('tag_name must not be None')

        return self._post(
            self._url('/namespaces/{0}/datasets/{1}/tags/{2}',
                      namespace_name, dataset_name, tag_name)
        )

    def tag_dataset_field(self, namespace_name, dataset_name, field_name,
                          tag_name):
        Utils.check_name_length(namespace_name, 'namespace_name')
        Utils.check_name_length(dataset_name, 'dataset_name')
        Utils.check_name_length(field_name, 'field_name')
        Utils.check_name_length(tag_name, 'tag_name')

        return self._post(
            self._url('/namespaces/{0}/datasets/{1}/fields/{2}/tags/{3}',
                      namespace_name, dataset_name, field_name, tag_name)
        )

    # Job API
    def create_job(self, namespace_name, job_name, job_type, location=None,
                   input_dataset=None, output_dataset=None, description=None,
                   context=None, run_id=None):
        Utils.check_name_length(namespace_name, 'namespace_name')
        Utils.check_name_length(job_name, 'job_name')
        Utils.is_instance_of(job_type, JobType)

        payload = {
            'inputs': input_dataset or [],
            'outputs': output_dataset or [],
            'type': job_type.name
        }

        if run_id:
            payload['runId'] = run_id

        if context:
            payload['context'] = context

        if location:
            payload['location'] = location

        if description:
            payload['description'] = description

        return self._put(
            self._url('/namespaces/{0}/jobs/{1}', namespace_name, job_name),
            payload=payload
        )

    def get_job(self, namespace_name, job_name):
        Utils.check_name_length(namespace_name, 'namespace_name')
        Utils.check_name_length(job_name, 'job_name')

        return self._get(
            self._url('/namespaces/{0}/jobs/{1}', namespace_name, job_name)
        )

    def list_jobs(self, namespace_name, limit=None, offset=None):
        Utils.check_name_length(namespace_name, 'namespace_name')

        return self._get(
            self._url('/namespaces/{0}/jobs', namespace_name),
            params={
                'limit': limit,
                'offset': offset
            }
        )

    def create_job_run(self, namespace_name, job_name, run_id=None,
                       nominal_start_time=None,
                       nominal_end_time=None, run_args=None,
                       mark_as_running=False):
        Utils.check_name_length(namespace_name, 'namespace_name')
        Utils.check_name_length(job_name, 'job_name')

        payload = {}

        if run_id:
            payload['id'] = run_id

        if nominal_start_time:
            payload['nominalStartTime'] = nominal_start_time

        if nominal_end_time:
            payload['nominalEndTime'] = nominal_end_time

        if run_args:
            payload['args'] = run_args

        response = self._post(
            self._url('/namespaces/{0}/jobs/{1}/runs',
                      namespace_name, job_name),
            payload=payload)

        if mark_as_running:
            response = self.mark_job_run_as_started(run_id)

        return response

    def list_job_runs(self, namespace_name, job_name, limit=None,
                      offset=None):
        Utils.check_name_length(namespace_name, 'namespace_name')
        Utils.check_name_length(job_name, 'job_name')

        return self._get(
            self._url(
                '/namespaces/{0}/jobs/{1}/runs',
                namespace_name,
                job_name),
            params={
                'limit': limit,
                'offset': offset
            }
        )

    def get_job_run(self, run_id):
        self._is_valid_uuid(run_id, 'run_id')

        return self._get(self._url('/jobs/runs/{0}', run_id))

    def mark_job_run_as_started(self, run_id, action_at=None):
        return self.__mark_job_run_as(run_id, 'start', action_at)

    def mark_job_run_as_completed(self, run_id, action_at=None):
        return self.__mark_job_run_as(run_id, 'complete', action_at)

    def mark_job_run_as_failed(self, run_id, action_at=None):
        return self.__mark_job_run_as(run_id, 'fail', action_at)

    def mark_job_run_as_aborted(self, run_id, action_at=None):
        return self.__mark_job_run_as(run_id, 'abort', action_at)

    def list_tags(self, limit=None, offset=None):
        return self._get(
            self._url('/tags'),
            params={
                'limit': limit,
                'offset': offset
            }
        )

    def __mark_job_run_as(self, run_id, action, action_at=None):
        Utils.is_valid_uuid(run_id, 'run_id')

        return self._post(
            self._url('/jobs/runs/{0}/{1}?at={2}', run_id, action,
                      action_at if action_at else Utils.utc_now()), payload={}
        )

    # Common
    def _url(self, path, *args):
        encoded_args = [quote(arg.encode('utf-8'), safe='') for arg in args]
        return f'{self._api_base}{path.format(*encoded_args)}'

    def _post(self, url, payload, as_json=True):
        now_ms = Utils.now_ms()

        response = requests.post(
            url=url, headers=_HEADERS, json=payload, timeout=self._timeout)

        post_details = {}
        post_details['url'] = url
        post_details['http_method'] = 'POST'
        post_details['http_headers'] = _HEADERS
        post_details['payload'] = payload
        post_details['duration_ms'] = (self._now_ms() - now_ms)

        log.info(post_details)

        return self._response(response, as_json)

    def _put(self, url, payload=None, as_json=True):
        now_ms = Utils.now_ms()

        response = requests.put(
            url=url, headers=_HEADERS, json=payload, timeout=self._timeout)

        put_details = {}
        put_details['url'] = url
        put_details['http_method'] = 'POST'
        put_details['http_headers'] = _HEADERS
        put_details['payload'] = payload
        put_details['duration_ms'] = (self._now_ms() - now_ms)

        log.info(put_details)

        return self._response(response, as_json)

    def _get(self, url, params=None, as_json=True):
        now_ms = Utils.now_ms()

        response = requests.get(
            url, params=params, headers=_HEADERS, timeout=self._timeout)

        get_details = {}
        get_details['url'] = url
        get_details['http_method'] = 'POST'
        get_details['http_headers'] = _HEADERS
        get_details['payload'] = params
        get_details['duration_ms'] = (self._now_ms() - now_ms)

        log.info(get_details)

        return self._response(response, as_json)

    def _response(self, response, as_json):
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            self._raise_api_error(e)

        return response.json() if as_json else response.text

    def _raise_api_error(self, e):
        # TODO: https://github.com/MarquezProject/marquez-python/issues/55
        raise errors.APIError() from e



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import uuid
from datetime import datetime

import pytz
import time
from pyrfc3339 import generate

from marquez_client.models import (DatasetFieldType, DatasetType)

log = logging.getLogger(__name__)


class Utils(object):
    def make_field(name, data_type, description=None):
        if isinstance(data_type, str):
            if not DatasetFieldType.__members__.__contains__(data_type):
                raise ValueError(f'Invalid field type: {data_type}')
        elif isinstance(data_type, DatasetFieldType):
            data_type = data_type.name
        else:
            raise ValueError('data_type must be a str or a DatasetFieldType')

        DatasetType.__members__.get(data_type)
        field = {
            'name': name,
            'type': data_type
        }
        if description:
            field['description'] = description
        return field

    @staticmethod
    def to_seconds(timeout_ms):
        return float(timeout_ms) / 1000.0

    @staticmethod
    def is_none(variable_value, variable_name):
        if not variable_value:
            raise ValueError(f"{variable_name} must not be None")

    @staticmethod
    def check_name_length(variable_value, variable_name):
        Utils.is_none(variable_value, variable_name)

        # ['namespace_name', 'owner_name', 'source_name'] <= 64
        # ['dataset_name', 'field_name', 'job_name', 'tag_name'] <= 255
        if variable_name in ['namespace_name', 'owner_name', 'source_name']:
            if len(variable_value) > 64:
                raise ValueError(f"{variable_name} length is"
                                 f" {len(variable_value)}, must be <= 64")
        else:
            if len(variable_value) > 255:
                raise ValueError(f"{variable_name} length is"
                                 f" {len(variable_value)}, must be <= 255")

    @staticmethod
    def is_valid_uuid(variable_value, variable_name):
        Utils.is_none(variable_value, variable_name)

        try:
            uuid.UUID(str(variable_value))
        except ValueError:
            raise ValueError(f"{variable_name} must be a valid UUID")

    @staticmethod
    def is_instance_of(variable_value, variable_enum_type):
        if not isinstance(variable_value, variable_enum_type):
            raise ValueError(f"{variable_value} must be an instance"
                             f" of {variable_enum_type}")

    @staticmethod
    def is_valid_connection_url(connection_url):
        Utils.is_none(connection_url, 'connection_url')

    @staticmethod
    def now_ms():
        return int(round(time.time() * 1000))

    @staticmethod
    def utc_now():
        return str(generate(datetime.utcnow().replace(tzinfo=pytz.utc),
                            microseconds=True))

    @staticmethod
    def get_json(file):
        with open(file) as json_file:
            return json.load(json_file)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
    Marquez

    Marquez is an open source **metadata service** for the **collection**, **aggregation**, and **visualization** of a data ecosystem's metadata.  # noqa: E501
"""
import codecs
import os

from setuptools import find_packages, setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('VERSION'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


NAME = "marquez-python"

# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools

setup(
    name=NAME,
    python_requires='>3.5.0',
    version=get_version('marquez_client/version.py'),
    description="Marquez Python Client",
    author_email="",
    url="",
    keywords=["Marquez"],
    packages=find_packages(),
    install_requires=[
        "requests",
        "six",
        "pyrfc3339"
    ],
    include_package_data=True,
    long_description="""\
    Marquez-Python is an open source library for building clients that
    interact with a running Marquez instance.
    """
)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from marquez_client import MarquezWriteOnlyClient


# TODO: Method _path() is internal, but adding a test for now.
#  We need to add tests for write only client that mocks the backend
def test_path():
    path = MarquezWriteOnlyClient._path('/test/{0}', '1')

    assert path == '/api/v1/test/1'



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

import marquez_client
from marquez_client.constants import (DEFAULT_MARQUEZ_BACKEND,
                                      DEFAULT_MARQUEZ_URL,
                                      DEFAULT_MARQUEZ_FILE,
                                      DEFAULT_TIMEOUT_MS)
from marquez_client.file_backend import FileBackend
from marquez_client.http_backend import HttpBackend
from marquez_client.log_backend import LogBackend
from marquez_client.utils import Utils

log = logging.getLogger(__name__)


# Marquez Clients
class Clients(object):
    def __init__(self):
        log.debug("Clients.init")

    @staticmethod
    def new_client():
        url = os.environ.get('MARQUEZ_URL', DEFAULT_MARQUEZ_URL)
        return marquez_client.MarquezClient(url)

    @staticmethod
    def new_write_only_client():
        return marquez_client.MarquezWriteOnlyClient(Clients.from_env())

    @staticmethod
    def from_env():
        backend_env = \
            os.environ.get('MARQUEZ_BACKEND', DEFAULT_MARQUEZ_BACKEND)

        if backend_env == 'http':
            url = os.environ.get('MARQUEZ_URL', DEFAULT_MARQUEZ_URL)
            timeout = Utils.to_seconds(
                os.environ.get('MARQUEZ_TIMEOUT_MS', DEFAULT_TIMEOUT_MS))
            return HttpBackend(url, timeout)
        elif backend_env == 'file':
            file = os.environ.get('MARQUEZ_FILE', DEFAULT_MARQUEZ_FILE)
            return FileBackend(file)
        elif backend_env == 'log':
            return LogBackend()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import logging.config
import unittest
import uuid

import mock
import yaml

import marquez_client
from marquez_client.models import (DatasetType, JobType, RunState)
from marquez_client.utils import Utils

_NAMESPACE = "my-namespace"
log = logging.getLogger(__name__)


class TestMarquezClient(unittest.TestCase):
    def setUp(self):
        log.debug("TestMarquezClient.setup(): ")

        with open('tests/logConfig.yaml', 'rt') as file:
            yamlConfig = yaml.safe_load(file.read())
            logging.config.dictConfig(yamlConfig)
            log.info("loaded logConfig.yaml")

        self.client = \
            marquez_client.client.MarquezClient("http://localhost:5000")
        log.info("created marquez_client.")

    @mock.patch("marquez_client.client.MarquezClient._put")
    def test_create_namespace(self, mock_put):
        owner_name = "me"
        description = "my namespace for testing."

        mock_put.return_value = {
            "name": _NAMESPACE,
            "ownerName": owner_name,
            "description": description
        }

        response = self.client.create_namespace(
            _NAMESPACE, owner_name, description)

        assert _NAMESPACE == str(response['name'])
        assert owner_name == str(response['ownerName'])
        assert description == str(response['description'])

    @mock.patch("marquez_client.client.MarquezClient._get")
    def test_get_namespace(self, mock_get):
        mock_get.return_value = {
            "name": _NAMESPACE
        }

        response = self.client.get_namespace(_NAMESPACE)

        assert _NAMESPACE == str(response['name'])

    @mock.patch("marquez_client.client.MarquezClient._get")
    def test_list_namespace(self, mock_get):
        mock_get.return_value = {
            "namespaces": [
                {
                    "name": _NAMESPACE,
                }
            ]
        }

        response = self.client.list_namespaces(limit=10, offset=0)

        assert _NAMESPACE == str(response['namespaces'][0]['name'])

    @mock.patch("marquez_client.client.MarquezClient._put")
    def test_create_dataset(self, mock_put):
        dataset_name = "my-dataset"
        description = "My dataset for testing."

        fields = [
            {
                "name": "flight_id",
                "type": "INTEGER",
                "description": "flight id"
            },
            {
                "name": "flight_name",
                "type": "VARCHAR",
                "description": "flight name"
            },
            {
                "name": "flight_date",
                "type": "TIMESTAMP",
                "description": "flight date"
            }
        ]

        mock_put.return_value = {
            'id': {
                'namespace': 'my-namespace',
                'name': 'my-dataset'
            },
            'type': 'DB_TABLE',
            'name': 'my-dataset',
            'physicalName': 'public.mytable',
            'createdAt': '2020-08-12T05:46:31.172877Z',
            'updatedAt': '2020-08-12T05:46:31.184934Z',
            'namespace': 'my-namespace',
            'sourceName': 'mydb',
            'fields': [
                {
                    'name': 'my_date',
                    'type': 'TIMESTAMP',
                    'description': 'my date'
                },
                {
                    'name': 'my_id',
                    'type': 'INTEGER',
                    'description': 'my id'
                },
                {
                    'name': 'my_name',
                    'type': 'VARCHAR',
                    'description': 'my name'
                }
            ],
            'tags': [],
            'lastModifiedAt': None,
            'description': 'My dataset for testing.'
        }

        response = self.client.create_dataset(
            namespace_name=_NAMESPACE,
            dataset_name=dataset_name,
            dataset_type=DatasetType.DB_TABLE,
            physical_name=dataset_name,
            source_name='my-source',
            description=description,
            run_id=None,
            schema_location=None,
            fields=fields,
            tags=None
        )

        assert str(response['description']) == description
        assert str(response['name']) == dataset_name

    @mock.patch("marquez_client.client.MarquezClient._get")
    def test_get_dataset(self, mock_get):
        dataset_name = "my-dataset"

        mock_get.return_value = {
            'id': {
                'namespace': 'my-namespace',
                'name': 'my-dataset'
            },
            'name': 'my-dataset',
        }

        response = self.client.get_dataset(
            namespace_name=_NAMESPACE,
            dataset_name=dataset_name)

        assert str(response['name']) == dataset_name

    @mock.patch("marquez_client.client.MarquezClient._put")
    def test_create_datasource(self, mock_put):
        source_name = "flight_schedules_db"
        source_type = 'POSTGRESQL'
        source_url = "jdbc:postgresql://localhost:5432/test?" \
                     "user=fred&password=secret&ssl=true"
        description = "PostgreSQL - flight schedules database"

        mock_put.return_value = {
            "type": "POSTGRESQL",
            "name": "flight_schedules_db",
            "createdAt": "2020-08-11T20:15:04.603158Z",
            "updatedAt": "2020-08-11T20:15:04.603158Z",
            "connectionUrl": "jdbc:postgresql://localhost:5432/"
                             "test?user=fred&password=secret&ssl=true",
            "description": "PostgreSQL - flight schedules database"
        }

        response = self.client.create_source(
            source_name=source_name,
            source_type=source_type,
            connection_url=source_url,
            description=description)

        assert response['name'] == source_name
        assert response['connectionUrl'] == source_url
        assert response['type'] == source_type

    @mock.patch("marquez_client.client.MarquezClient._put")
    def test_create_job(self, mock_put):
        job_name = "my-job"
        input_dataset = [
            {
                "namespace": "my-namespace",
                "name": "public.mytable"
            }
        ]
        output_dataset = {
            "namespace": "my-namespace",
            "name": "public.mytable"
        }

        location = "https://github.com/my-jobs/blob/" \
                   "07f3d2dfc8186cadae9146719e70294a4c7a8ee8"

        context = {
            "SQL": "SELECT * FROM public.mytable;"
        }

        mock_put.return_value = {
            "id": {
                "namespace": "my-namespace",
                "name": "my-job"
            },
            "type": "BATCH",
            "name": "my-job",
            "createdAt": "2020-08-12T07:30:55.321059Z",
            "updatedAt": "2020-08-12T07:30:55.333230Z",
            "namespace": "my-namespace",
            "inputs": [
                {
                    "namespace": "my-namespace",
                    "name": "public.mytable"
                }
            ],
            "outputs": [
                {
                    "namespace": "my-namespace",
                    "name": "public.mytable"
                }
            ],
            "location": "https://github.com/my-jobs/blob/"
                        "07f3d2dfc8186cadae9146719e70294a4c7a8ee8",
            "context": {
                "SQL": "SELECT * FROM public.mytable;"
            },
            "description": "My first job.",
            "latestRun": None
        }

        response = self.client.create_job(
            namespace_name=_NAMESPACE,
            job_name=job_name,
            job_type=JobType.BATCH,
            location=location,
            input_dataset=input_dataset,
            output_dataset=output_dataset,
            context=context
        )

        assert str(response['id']) is not None
        assert str(response['location']) == location

    @mock.patch("marquez_client.client.MarquezClient._post")
    def test_create_job_run(self, mock_post):
        run_id = str(uuid.uuid4())
        action_at = Utils.utc_now()

        job_name = "my-job"
        run_args = {
            "email": "me@mycorp.com",
            "emailOnFailure": "true",
            "emailOnRetry": "true",
            "retries": "1"
        }

        mock_post.return_value = {
            'id': f'{run_id}',
            'createdAt': '2020-08-12T22:33:02.787228Z',
            'updatedAt': '2020-08-12T22:33:02.787228Z',
            'nominalStartTime': None,
            'nominalEndTime': None,
            'state': 'NEW',
            'startedAt': f'{action_at}',
            'endedAt': None,
            'durationMs': None,
            'args': {
                "email": "me@mycorp.com",
                "emailOnFailure": "true",
                "emailOnRetry": "true",
                "retries": "1"
            }
        }

        response = self.client.create_job_run(
            namespace_name=_NAMESPACE,
            job_name=job_name,
            run_id=run_id,
            nominal_start_time=None,
            nominal_end_time=None,
            run_args=run_args,
            mark_as_running=True
        )

        assert response['id'] is not None
        assert str(response['args']) == str(run_args)
        assert str(response['startedAt']) == action_at

    @mock.patch("marquez_client.client.MarquezClient._post")
    def test_mark_job_run_as_start(self, mock_post):
        run_id = str(uuid.uuid4())
        action_at = Utils.utc_now()

        mock_post.return_value = {
            'id': f'{run_id}',
            'createdAt': '2020-08-12T22:36:50.739951Z',
            'updatedAt': '2020-08-13T17:56:39.516802Z',
            'nominalStartTime': None,
            'nominalEndTime': None,
            'state': 'RUNNING',
            'startedAt': f'{action_at}',
            'endedAt': None,
            'durationMs': None,
            'args': {}
        }

        response = self.client.mark_job_run_as_started(
            run_id=run_id, action_at=action_at)

        assert str(response['id']) == run_id
        assert str(response['state']) == RunState.RUNNING.value
        assert str(response['startedAt']) == action_at

    @mock.patch("marquez_client.client.MarquezClient._post")
    def test_mark_job_run_as_completed(self, mock_post):
        run_id = str(uuid.uuid4())
        action_at = Utils.utc_now()

        mock_post.return_value = {
            'id': f'{run_id}',
            'createdAt': '2020-08-12T22:36:50.739951Z',
            'updatedAt': '2020-08-13T17:56:39.516802Z',
            'nominalStartTime': None,
            'nominalEndTime': None,
            'state': 'COMPLETED',
            # 'startedAt': '2020-08-13T17:56:39.516802Z',
            'startedAt': f'{action_at}',
            'endedAt': None,
            'durationMs': None,
            'args': {}
        }

        response = self.client.mark_job_run_as_completed(
            run_id=run_id, action_at=action_at)

        assert str(response['id']) == run_id
        assert str(response['state']) == RunState.COMPLETED.value
        assert str(response['startedAt']) == action_at

    @mock.patch("marquez_client.client.MarquezClient._post")
    def test_mark_job_run_as_failed(self, mock_post):
        run_id = str(uuid.uuid4())
        action_at = Utils.utc_now()

        mock_post.return_value = {
            'id': f'{run_id}',
            'createdAt': '2020-08-12T22:36:50.739951Z',
            'updatedAt': '2020-08-13T17:56:39.516802Z',
            'nominalStartTime': None,
            'nominalEndTime': None,
            'state': 'FAILED',
            'startedAt': f'{action_at}',
            'endedAt': None,
            'durationMs': None,
            'args': {}
        }

        response = self.client.mark_job_run_as_failed(
            run_id=run_id, action_at=action_at)

        assert str(response['id']) == run_id
        assert str(response['state']) == RunState.FAILED.value
        assert str(response['startedAt']) == action_at

    @mock.patch("marquez_client.client.MarquezClient._post")
    def test_mark_job_run_as_aborted(self, mock_post):
        run_id = str(uuid.uuid4())
        action_at = Utils.utc_now()

        mock_post.return_value = {
            'id': f'{run_id}',
            'createdAt': '2020-08-12T22:36:50.739951Z',
            'updatedAt': '2020-08-13T17:56:39.516802Z',
            'nominalStartTime': None,
            'nominalEndTime': None,
            'state': 'ABORTED',
            'startedAt': f'{action_at}',
            'endedAt': None,
            'durationMs': None,
            'args': {}
        }

        response = self.client.mark_job_run_as_aborted(
            run_id=run_id, action_at=action_at)

        assert str(response['id']) == run_id
        assert str(response['state']) == RunState.ABORTED.value
        assert str(response['startedAt']) == action_at


if __name__ == '__main__':
    unittest.main()
